		var INPUT_ELEMENT_CTRL_TYPES = ["text", "password", "radio", "checkbox"];
	
		function asList (nodeList, iid) {
			var res = new Array();
			if (nodeList != null ) {
				for ( var index=0; lt( index, nodeList.length); index++ ) {
					var elem = nodeList.item(index);
					if ( iid != null ) { 
						elem = elem.queryInterface(iid);
					}
					res[ res.length ] = elem;
				}
			} 
			return res;
		}
	     
// pass in a string to verify it is valid alt-text
function isValidAltText (altText) {
	var result = false;
	if ( altText != null ) {
		var pattern = new RegExp("\/?([\w\-]\/?)+\.\w+");
		result = and( isNonEmptyString(altText), !pattern.test(String(altText)));
	}	
    return result;
}

		function hasTextualContent (element) {
			var res = isNonEmptyTextNode( element );
			if ( !res ) {
				var children = asList(element.childNodes);
				for ( var i=0; and(!res, lt( i, children.length)); i++ ) {
					res = hasTextualContent(children[i]);
				}
			}
			return res;
		}
	   
		function isNonEmptyTextNode ( node ) {
			return and(node.nodeType == nsIDOMNode.TEXT_NODE, isNonEmptyString( node.nodeValue ));
		}
			     
		function hasNoFrameSourceError ( src ) {
			var extList = ["jpg","gif","bmp","tif"];
			return !(hasExtension( src, extList ) );    
		}
	       
		function isLabeled ( element ) {
			var res = false;
			var id = null;
			
			if ( element != null ) {
				id = element.id;
			}
	
			if (element != null && id != null) {
				if (isNonEmptyString(id)) {
					var labels = asList(element.ownerDocument.getElementsByTagName('LABEL'), nsIDOMHTMLLabelElement.NS_IDOMHTMLLABELELEMENT_IID);
					for ( var i=0; i < labels.length && !res; i++ ) {
						label = labels[i];
						res = ( String(id) == label.htmlFor );
					}
				}
			}
			
			return res;
		}
	     
		function isLayoutTable (tableElement) {
			var tables = asList(tableElement.getElementsByTagName('TABLE'));
			var hasChildTables = gt( tables.length, 0);
	       var rows = asList(tableElement.rows, nsIDOMHTMLTableRowElement.NS_IDOMHTMLTABLEROWELEMENT_IID);
			var hasTextCell = false;
			
			rowIteration:
			for ( var i=0; lt( i, rows.length); i++ ) {
				var row = rows[i];
				var cells = asList(row.cells, nsIDOMHTMLTableCellElement.NS_IDOMHTMLTABLECELLELEMENT_IID);
				for ( var j=0; lt( j, cells.length); j++ ) {
					if (hasTextualContent(cells[j])) {
						hasTextCell = true;
						break rowIteration;
					}
				}
			}
			
			return or(hasChildTables, !hasTextCell);
		}
	
		function hasSufficientCellMarkup (table) {
		    var rows = asList(table.rows, nsIDOMHTMLTableRowElement.NS_IDOMHTMLTABLEROWELEMENT_IID);
		    var result = false;
		    if (isSimpleDataTable(rows)) {
		        result = hasTableHeaders(rows);
		    } else {
		    	// complex data table
		        // result = and(hasTableHeaders(rows), hasCellHeaders(rows));
		    }
		    return result;
		}
		
		function isSimpleDataTable (rows) {
            var res = true;
            
            rowIteration:
            for ( var i=0; lt( i, rows.length); i++ ) {
                var cells = asList(rows[i].cells, nsIDOMHTMLTableCellElement.NS_IDOMHTMLTABLECELLELEMENT_IID);
                for ( var j=0; lt( j, cells.length); j++ ) {
                    var cell = cells[j];
                    if (eq(cell.nodeName, 'TH')) {
	                    if (gt(j, 0)) {
	                        // TH occurs past first column and not in first row
	                        // so cannot be simple data table
	                        res = false;
	                        break rowIteration;
	                    }
	                } else { // assume TD element
                        res = hasTextualContent(cell);
                        if (!res) {
                            break rowIteration;
                        }
                    }
                }
            }
            
            return res;
		}
		
		function hasTableHeaders (tableRows) {
			var result = true;
			var startRow = 1;
			
			// first row should consist entirely of TH elements
			if (eq(tableRows.length, 0)) {
			 return false;
			}
            var cells = asList(tableRows[0].cells, nsIDOMHTMLTableCellElement.NS_IDOMHTMLTABLECELLELEMENT_IID);
            for ( var j=0; and(result, lt(      j, cells.length)); j++ ) {
                result = (cells[j].nodeName == 'TH');
            }                 
			
            // second row may consist entirely of TH elements if
            // this is a complex data table; assumption is that header are
            // never nested more than two levels
            var thIndex = -1, tdIndex = -1;
            if (gt(tableRows.length, 1)) {
                cells = asList(tableRows[1].cells, nsIDOMHTMLTableCellElement.NS_IDOMHTMLTABLECELLELEMENT_IID);
                for ( var jj=0; and(result, lt(      jj, cells.length)); j++ ) {
                    if (cells[jj].nodeName == 'TH') {
                        thIndex = jj;
                    } else if (cells[jj].nodeName == 'TD') {
                        tdIndex = jj;
                    }
                }
                if (tdIndex == -1) {
                    // all TH elements so this is complex data table
                	++startRow;                  
                } else if (thIndex == -1) {
                    // no headers in row so fails test
                    result = false;
                } else if (lt(tdIndex, thIndex)) {
                    // TH found after TD so table fails
                    result = false;
                }
            }
                
			// now, simply verify that each row has at least one header element
			for ( var i=startRow; and(result, lt( i, tableRows.length)); i++ ) {
				cells = asList(tableRows[i].cells, nsIDOMHTMLTableCellElement.NS_IDOMHTMLTABLECELLELEMENT_IID);
				for ( var jjj=0; lt(jjj, cells.length); j++ ) {
	                if (cells[jjj].nodeName == 'TH') {
	                    thIndex = jjj;
	                } else if (cells[jjj].nodeName == 'TD') {
	                    tdIndex = jjj;
	                }
				}                 

	            if (thIndex == -1) {
	                // no headers in row so fails test
	                result = false;
	            } else if (lt(tdIndex, thIndex)) {
	                // TH found after TD so table fails
	                result = false;
	            }
			}

            return result;
		}
	
		function isPlaceHolderImage (imgElement) {
			var imgSrc = imgElement.src.toLowerCase();
			var isLargeImage = and( gt(imgElement.width, 10), gt(imgElement.height, 10) );
			var hasPlaceholderName = or( gt(imgSrc.indexOf("spacer"), -1), gt(imgSrc.indexOf("blank"), -1) );
			var result = or( hasPlaceholderName , !isLargeImage );
			return result;
		}
	 
		function hasSkipToMainContent ( element ) {
			var res = false;
			var tags = asList( element.getElementsByTagName("A") , nsIDOMHTMLAnchorElement.NS_IDOMHTMLANCHORELEMENT_IID);
			
			for ( var i=0; lt( i, tags.length ); i++  ) {
				var tag = tags[i];
				if (isSkipAnchor( tag )) {
					res = true;
					break;
				}
			}
			
			return res;
		}
	
		function isSkipAnchor (anchor) {
			var foundPound = false;
			var foundPoundNavSkip = false;
			var foundSkipMsg = false;
			var result = false;
			var fullHref = anchor.href;
			var splitHref = null;
		
			// determine if we have a valid href for skip anchor
			if ( fullHref != null ) {
				splitHref = fullHref.split('#');
				foundPound = ( splitHref.length == 2 );		
			}
		
			if ( foundPound ) {
				if ( splitHref[1].toLowerCase().indexOf('navskip') > -1 ) {
					foundPoundNavSkip = true;
				}  
				var children = asList(anchor.childNodes);
				for ( var i=0; lt( i, children.length); i++ ) {
					var child = children[i];
					var text = '';
					if ( child.nodeType == nsIDOMNode.TEXT_NODE ) {
						text =  child.nodeValue;
					}
					if ( and( child.nodeType == nsIDOMNode.ELEMENT_NODE, child.nodeName == 'IMG')) {
						text = child.nodeValue;
					}
					if ( text != null ) {
						if ( gt( text.toLowerCase().indexOf('skip'), -1) ) {  
							foundSkipMsg = true;
							break;
						}
					}
				}
				result = ( or( foundPoundNavSkip, ( and( foundPound, foundSkipMsg ) ) ));
			}
			return result;
		}
	      
		function isInputLabeled ( element ) {
			var inputType = element.type.toLowerCase();
	     	var isValidInputType = gt( INPUT_ELEMENT_CTRL_TYPES.join().indexOf(inputType), -1);
	     	return or ( !isValidInputType, isLabeled(element) );
		}
